#include <sys/types.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <netinet/in.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#define	QLEN		   5	/* maximum connection queue length	*/
#define	BUFSIZE		4096

extern int	errno;

void	reaper(int);
int	proxyechod(int client_fd);
int	errexit(const char *format, ...);
int	passiveTCP(const char *service, int qlen);
int     connectTCP(const char *host, const char *service);

char* real_proxy = (char*)0;
char* real_port = (char*)0;


int main(int argc, char *argv[]){
  char	*service = "http";	/* service name or port number	*/
  struct	sockaddr_in fsin;	/* the address of a client	*/
  int	alen;			/* length of client's address	*/
  int	msock;			/* master server socket		*/
  int	ssock;			/* slave server socket		*/

  switch (argc) {
  case 4:
    real_port = argv[3];
  case 3:
    service = argv[2];
  case 2:
    real_proxy = argv[1];
    break;
  default:
    errexit("usage: proxy <realProxy> [server port] [realProxy port]\n");
  }

  msock = passiveTCP(service, QLEN);

  (void) signal(SIGCHLD, reaper);

  while (1) {
    alen = sizeof(fsin);
    ssock = accept(msock, (struct sockaddr *)&fsin, &alen);
    if (ssock < 0) {
      if (errno == EINTR)
	continue;
      errexit("accept: %s\n", strerror(errno));
    }
    switch (fork()) {
    case 0:		/* child */
      (void) close(msock);
      exit(proxyechod(ssock));
    default:	/* parent */
      (void) close(ssock);
      break;
    case -1:
      errexit("fork: %s\n", strerror(errno));
    }
  }
}

char  buffer[BUFSIZE];
int   buf_pos = 0;
char* keyword;

void sock_printf(int fd, const char* format, ...){
  va_list	args;
  char temp[BUFSIZE];
  int size;
  
  va_start(args, format);
  size = vsprintf(temp, format, args);
  va_end(args);

  printf(temp);
  fflush(stdout);
  write(fd, temp, size);
}

void print_buf(char* buf, int upto){
  int i;
  for(i=0; i<upto; i++)
    if(iscntrl(buf[i]) || isprint(buf[i]))
      fputc(buf[i], stdout);
}

void transfer(int proxy_fd, int client_fd){
  fd_set rfds;
  int nfds = getdtablesize();
  struct timeval base = {30, 0};	/* timeout after 30 seconds */
  struct timeval cur = base;
  int retval;
  char buf[BUFSIZE];
  int bytes_read = 0;
  int echo_fd = fileno(stdout);

  while(1){
    FD_ZERO(&rfds);
    FD_SET(proxy_fd, &rfds);
    FD_SET(client_fd, &rfds);

    retval = select(nfds, &rfds, NULL, NULL, &cur);
    cur = base;
    if(!retval)
      break;
    
    if(FD_ISSET(proxy_fd, &rfds)){
      bytes_read = read(proxy_fd, buf, BUFSIZE);

      if(bytes_read <= 0)
	break;

      write(echo_fd, buf, bytes_read);
      
      if(write(client_fd, buf, bytes_read) <= 0)
	break;
    }
      
    if(FD_ISSET(client_fd, &rfds)){
      bytes_read = read(client_fd, buf, BUFSIZE);

      if(bytes_read <= 0)
	break;

      write(echo_fd, buf, bytes_read);

      if(write(proxy_fd, buf, bytes_read) <= 0)
	break;
    } 
  }

  printf("session complete\n");
  fflush(stdout);

  close(proxy_fd);
  close(client_fd);
}

int proxyechod(int client_fd){
  int proxy_fd = connectTCP(real_proxy, real_port);

  transfer(proxy_fd, client_fd);

  return 0;
}

/*------------------------------------------------------------------------
 * reaper - clean up zombie children
 *------------------------------------------------------------------------
 */
/*ARGSUSED*/
void reaper(int sig){
  int	status;

  while (wait3(&status, WNOHANG, (struct rusage *)0) >= 0)
    /* empty */;
}
